@@ -1,5 +1,5 @@
 /*
- * DVDRW format 6.1 by Andy Polyakov <appro@fy.chalmers.se>.
+ * DVDRW format 7.0 by Andy Polyakov <appro@fy.chalmers.se>.
  *
  * Use-it-on-your-own-risk, GPL bless...
  *
@@ -63,6 +63,9 @@
  * 6.1:
  * -  localization;
  * - Treat only x73xx OPC errors as fatal;
+ * 7.0:
+ * - Blu-ray Disc support;
+ * - Mac OS X 10>=2 support;
  */
 #include <stdio.h>
 #include <stdlib.h>
@@ -83,7 +86,7 @@
 
 static void usage (char *prog)
 { fprintf (stderr,"- usage: %s [-force[=full]] [-lead-out|-blank[=full]]\n"
-		  "         [-ssa[=none|default|max]] /dev/dvd\n",prog),
+		  "         [-ssa[=none|default|max|XXXm]] /dev/dvd\n",prog),
   exit(1);
 }
 
@@ -149,15 +152,12 @@ extern "C" void alarm_handler (int no)
 #endif
 
 int main (int argc, char *argv[])
-{ Scsi_Command	cmd;
-  unsigned char	formats[260],dinfo[32],inq[128];
+{ unsigned char	formats[260],dinfo[32],inq[128];
   char		*dev=NULL,*p;
   unsigned int	capacity,lead_out,mmc_profile,err;
-  int		force=0,full=0,compat=0,blank=0,ssa=0,do_opc=0,gui=0,len,i;
-#if defined(__unix) || defined(__unix__)
-  pid_t		pid;
-#endif
-
+  int		len,i;
+  int		force=0,full=0,compat=0,blank=0,ssa=0,do_opc=0,gui=0,
+		blu_ray=0,not_pow=0;
 #ifdef _WIN32
   DWORD		ppid;
   char		filename[MAX_PATH],*progname;
@@ -208,10 +208,49 @@ int main (int argc, char *argv[])
     progname = strrchr(filename,'\\');
     if (progname)	argv[0] = progname+1;
     else		argv[0] = filename;
+#elif defined(__unix) || defined(__unix__)
+  pid_t		pid;
+
+    { int fd;
+      char *s;
+
+	if ((fd=mkstemp (s=strdup("/tmp/dvd+rw-format.XXXXXX"))) < 0)
+	    fprintf (stderr,":-( unable to mkstemp(\"%s\")",s),
+	    exit(1);
+
+	ftruncate(fd,sizeof(*progress));
+	unlink(s);
+
+	progress = (int *)mmap(NULL,sizeof(*progress),PROT_READ|PROT_WRITE,
+				MAP_SHARED,fd,0);
+	close (fd);
+	if (progress == MAP_FAILED)
+            perror (":-( unable to mmap anonymously"),
+	    exit(1);
+    }
+    *progress = 0;
+
+    if ((pid=fork()) == (pid_t)-1)
+	perror (":-( unable to fork()"),
+	exit(1);
+
+    if (pid)
+    { struct sigaction sa;
+
+	sigaction (SIGALRM,NULL,&sa);
+	sa.sa_flags &= ~SA_RESETHAND;
+	sa.sa_flags |= SA_RESTART;
+	sa.sa_handler = alarm_handler;
+	sigaction (SIGALRM,&sa,NULL);
+	alarm(1);
+	while ((waitpid(pid,&i,0) != pid) && !WIFEXITED(i)) ;
+	if (WEXITSTATUS(i) == 0) fprintf (stderr,"\n");
+	exit (0);
+    }
 #endif
 
-    fprintf (stderr,"* DVD%sRW/-RAM format utility by <appro@fy.chalmers.se>, "
-		    "version 6.1.\n",plusminus_locale());
+    fprintf (stderr,"* BD/DVD%sRW/-RAM format utility by <appro@fy.chalmers.se>, "
+		    "version 7.0.\n",plusminus_locale());
 
     for (i=1;i<argc;i++) {
 	if (*argv[i] == '-')
@@ -226,13 +265,28 @@ int main (int argc, char *argv[])
 		if ((p=strchr(argv[i],'=')) && p[1]=='f') blank=0x10;
 		force=1;
 	    }
-	    else if (argv[i][1] == 's')	// -ssa
+	    else if (argv[i][1] == 'p')	// -pow
+	    {	not_pow=1;		// minus pow
+	    }
+	    else if (argv[i][1] == 's')	// -ssa|-spare
 	    {	force=ssa=1;
 		if ((p=strchr(argv[i],'=')))
 		{   if (p[1]=='n')	ssa=-1;	// =none
+		    else if (p[3]=='n') ssa=1;	// =min
 		    else if (p[1]=='d')	ssa=2;	// =default
 		    else if (p[3]=='x')	ssa=3;	// =max
-		    else		ssa=0;	// invalid
+		    else if (p[1]=='.' || (p[1]>='0' && p[1]<='9'))
+		    { char  *s=NULL;
+		      double a=strtod(p+1,&s);
+
+			if (s)
+			{   if      (*s=='g' || *s=='G')    a *= 1e9;
+			    else if (*s=='m' || *s=='M')    a *= 1e6;
+			    else if (*s=='k' || *s=='K')    a *= 1e3;
+			}
+			ssa = (int)(a/2e3);	// amount of 2K
+		    }
+		    else		ssa=0;	// invalid?
 		}
 	    }
 	    else if (argv[i][1] == 'g')	gui=1;
@@ -249,6 +303,8 @@ int main (int argc, char *argv[])
 
     if (dev==NULL) usage (argv[0]);
 
+    Scsi_Command	cmd;
+
     if (!cmd.associate(dev))
 	fprintf (stderr,":-( unable to open(\"%s\"): ",dev), perror (NULL),
 	exit(1);
@@ -271,11 +327,14 @@ int main (int argc, char *argv[])
 
     mmc_profile = formats[6]<<8|formats[7];
 
+    blu_ray = ((mmc_profile&0xF0)==0x40);
+
     if (mmc_profile!=0x1A && mmc_profile!=0x2A
 	&& mmc_profile!=0x14 && mmc_profile!=0x13
-	&& mmc_profile!=0x12)
+	&& mmc_profile!=0x12
+	&& !blu_ray)
 	fprintf (stderr,":-( mounted media doesn't appear to be "
-			"DVD%sRW or DVD-RAM\n",plusminus_locale()),
+			"DVD%sRW, DVD-RAM or Blu-ray\n",plusminus_locale()),
 	exit (1);
 
     /*
@@ -335,9 +394,6 @@ int main (int argc, char *argv[])
 		    }
 		}
 		break;
-	    case 0:
-		i=8;	// just grab the first descriptor?
-		break;
 	    case 1:	// first ssa
 		for (i=8;i<len;i+=8)
 		    if ((formats[4+i+4]>>2) == 0x01) break;
@@ -353,10 +409,170 @@ int main (int argc, char *argv[])
 		    }
 		}
 		break;
+	    default:
+		i=8;	// just grab the first descriptor?
+		break;
+	}
+    }
+    else if (mmc_profile==0x41)		// BD-R
+    { unsigned int   max,min,cap;
+      unsigned char *f;
+      int j;
+
+	switch (ssa)
+	{   case -1:	// no spare -> nothing to do
+	    case 0:
+		exit (0);
+		break;
+	    case 1:	// min spare <- max capacity
+		i=len;
+		for (max=0,j=8;j<len;j+=8)
+		{   f = formats+4+j;
+		    if ((f[4]>>2) == 0x32)
+		    {	cap = f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+			if (max < cap) max=cap,i=j;
+		    }
+		}
+		break;
+	    case 2:	// default ssa
+		i=8;	// just grab first descriptor
+		break;
+	    case 3:	// max, ~10GB, is too big to trust user
+		fprintf (stderr,"- -ssa=max is not supported for BD-R, "
+				"specify explicit size with -ssa=XXXG\n");
+		exit (1);
+	    default:
+		i=len;
+		for (max=0,min=0xffffffff,j=8;j<len;j+=8)
+		{   f = formats+4+j;
+		    if ((f[4]>>2) == 0x32)
+		    {	cap = f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+			if (max < cap) max=cap;
+			if (min > cap) min=cap;
+		    }
+		}
+		if (max==0) break;
+ 
+		// for simplicity adjust spare size according to DL(!) rules
+		ssa += 8192, ssa &= ~16383;	// i.e. in 32MB increments
+
+		f = formats+4;
+		capacity = f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+
+		cap = capacity - ssa;
+		// place it within given boundaries
+		if      (cap < min)	cap = min;
+		else if (cap > max)	cap = max;
+
+		i = 8;
+		f = formats+4+i;
+		f[0] = cap>>24;
+		f[1] = cap>>16;
+		f[2] = cap>>8;
+		f[3] = cap;
+		f[4] = 0x32<<2 | not_pow;
+		f[5] = 0;
+		f[6] = 0;
+		f[7] = 0;
+		break;
+	}
+	
+	if (i<len)
+	{   f = formats+4+i;
+	    f[4] &= ~3;		// it's either SRM+POW ...
+	    f[4] |= not_pow;	// ... or SRM-POW
+	}
+    }
+    else if (mmc_profile==0x43)		// BD-RE
+    { unsigned int   max,min,cap;
+      unsigned char *f;
+      int j;
+
+	switch (ssa)
+	{   case -1:	// no spare
+		for (i=8;i<len;i+=8)	// look for descriptor 0x31
+		    if ((formats [4+i+4]>>2) == 0x31) break;
+		break;
+	    case 0:
+		i = 8;
+		if ((formats[4+4]&3)==2)// same capacity for already formatted
+		{   f = formats+4+i;
+		    memcpy (f,formats+4,4);
+		    f[4] = 0x30<<2;
+		    f[5] = 0;
+		    f[6] = 0;
+		    f[7] = 0;
+		}
+		break;
+	    case 1:	// min spare <- max capacity
+		i=len;
+		for (max=0,j=8;j<len;j+=8)
+		{   f = formats+4+j;
+		    if ((f[4]>>2) == 0x30)
+		    {	cap = f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+			if (max < cap) max=cap,i=j;
+		    }
+		}
+		break;
+	    case 2:	// default ssa
+		i=8;	// just grab first descriptor
+		break;
+	    case 3:	// max spare <- min capacity
+		i=len;
+		for (min=0xffffffff,j=8;j<len;j+=8)
+		{   f = formats+4+j;
+		    if ((f[4]>>2) == 0x30)
+		    {	cap=f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+			if (min > cap) min=cap,i=j;
+		    }
+		}
+		break;
+	    default:
+		i=len;
+		capacity=0;
+		for (max=0,min=0xffffffff,j=8;j<len;j+=8)
+		{   f = formats+4+j;
+		    if ((f[4]>>2) == 0x30)
+		    {	cap = f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+			if (max < cap) max=cap;
+			if (min > cap) min=cap;
+		    }
+		    else if ((f[4]>>2) == 0x31)
+			capacity = f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+		}
+		if (max==0 || capacity==0) break;
+ 
+		// for simplicity adjust spare size according to DL(!) rules
+		ssa += 8192, ssa &= ~16383;	// i.e. in 32MB increments
+
+		cap = capacity - ssa;
+		// place it within given boundaries
+		if      (cap < min)	cap = min;
+		else if (cap > max)	cap = max;
+
+		i = 8;
+		f = formats+4+i;
+		f[0] = cap>>24;
+		f[1] = cap>>16;
+		f[2] = cap>>8;
+		f[3] = cap;
+		f[4] = 0x30<<2;
+		f[5] = 0;
+		f[6] = 0;
+		f[7] = 0;
+		break;
+	}
+
+	if (i<len)
+	{   f = formats+4+i;
+	    f[4] &= ~3;
+	    if ((f[4]>>2)==0x30)
+		f[4] |= full?2:3;	// "Full" or "Quick Certification"
 	}
     }
     else			// DVD-RW
     { int descr=full?0x10:0x15;
+
 	for (i=8;i<len;i+=8)	// look for "DVD-RW Quick" descriptor
 	    if ((formats [4+i+4]>>2) == descr) break;
 	if (descr==0x15 && i==len)
@@ -373,18 +589,29 @@ int main (int argc, char *argv[])
     }
 
     capacity = 0;
-    capacity |= formats[4+i+0], capacity <<= 8;
-    capacity |= formats[4+i+1], capacity <<= 8;
-    capacity |= formats[4+i+2], capacity <<= 8;
-    capacity |= formats[4+i+3];
+    if (blu_ray)
+    {	capacity |= formats[4+0], capacity <<= 8;
+	capacity |= formats[4+1], capacity <<= 8;
+	capacity |= formats[4+2], capacity <<= 8;
+	capacity |= formats[4+3];
+    }
+    else
+    {	capacity |= formats[4+i+0], capacity <<= 8;
+	capacity |= formats[4+i+1], capacity <<= 8;
+	capacity |= formats[4+i+2], capacity <<= 8;
+	capacity |= formats[4+i+3];
+    }
 
     if (mmc_profile==0x1A || mmc_profile==0x2A)	// DVD+RW
 	fprintf (stderr,"* %.1fGB DVD+RW media detected.\n",
 			2048.0*capacity/1e9);
-    else if (mmc_profile==0x12)	// DVD-RAM
+    else if (mmc_profile==0x12)			// DVD-RAM
 	fprintf (stderr,"* %.1fGB DVD-RAM media detected.\n",
 			2048.0*capacity/1e9);
-    else			// DVD-RW
+    else if (blu_ray)				// BD
+	fprintf (stderr,"* %.1fGB BD media detected.\n",
+			2048.0*capacity/1e9);
+    else					// DVD-RW
 	fprintf (stderr,"* %.1fGB DVD-RW media in %s mode detected.\n",
 			2048.0*capacity/1e9,
 			mmc_profile==0x13?"Restricted Overwrite":"Sequential");
@@ -432,6 +659,14 @@ int main (int argc, char *argv[])
 				"                to grow, eliminate, reset to default or\n"
 				"                maximize Supplementary Spare Area.\n",
 				argv[0]);
+	    else if (mmc_profile == 0x43)
+		fprintf (stderr,"- you have the option to re-run %s with:\n"
+				"  -format=full  to perform (lengthy) Full Certification;\n"
+				"  -ssa[=none|default|max|XXX[GM]]\n"
+				"                to eliminate or adjust Spare Area.\n",
+				argv[0]);
+	    else if (blu_ray)	// BD-R
+		fprintf (stderr,"- BD-R can be pre-formatted only once\n");
 	    else
 	    {	fprintf (stderr,"- you have the option to re-run %s with:\n",
 				argv[0]);
@@ -454,12 +689,17 @@ int main (int argc, char *argv[])
 			    "  to insist on -force option upon mode transition.\n");
 	    exit (0);
 	}
+	else if (mmc_profile==041 && !force)
+	{   fprintf (stderr,"- media is blank\n");
+	    fprintf (stderr,"- as BD-R can be pre-formance only once, I've chosen\n"
+			    "  to insist on -force option.\n");
+	}
 	force = 0;
     }
 
     if (((mmc_profile == 0x1A || mmc_profile == 0x2A) && blank)
 	|| (mmc_profile != 0x1A && compat)
-	|| (mmc_profile != 0x12 && ssa) )
+	|| (mmc_profile != 0x12 && mmc_profile != 0x43 && ssa) )
     {	fprintf (stderr,"- illegal command-line option for this media.\n");
 	goto offer_options;
     }
@@ -469,51 +709,6 @@ int main (int argc, char *argv[])
     }
 
 #if defined(__unix) || defined(__unix__)
-    { int fd;
-      char *s;
-
-	if ((fd=mkstemp (s=strdup("/tmp/dvd+rw.XXXXXX"))) < 0)
-	    fprintf (stderr,":-( unable to mkstemp(\"%s\")",s),
-	    exit(1);
-
-	ftruncate(fd,sizeof(*progress));
-	unlink(s);
-
-	progress = (int *)mmap(NULL,sizeof(*progress),PROT_READ|PROT_WRITE,
-				MAP_SHARED,fd,0);
-	close (fd);
-	if (progress == MAP_FAILED)
-            perror (":-( unable to mmap anonymously"),
-	    exit(1);
-    }
-    *progress = 0;
-
-    if ((pid=fork()) == (pid_t)-1)
-	perror (":-( unable to fork()"),
-	exit(1);
-
-    if (pid)
-    { struct sigaction sa;
-      const char *str;
-
-	cmd.~Scsi_Command();
-	if (compat && force)	str="relocating lead-out";
-	else if (blank)		str="blanking";
-	else			str="formatting";
-	if (gui)		gui_action=str;
-	else			fprintf (stderr,"* %s .",str);
-
-	sigaction (SIGALRM,NULL,&sa);
-	sa.sa_flags &= ~SA_RESETHAND;
-	sa.sa_flags |= SA_RESTART;
-	sa.sa_handler = alarm_handler;
-	sigaction (SIGALRM,&sa,NULL);
-	alarm(1);
-	while ((waitpid(pid,&i,0) != pid) && !WIFEXITED(i)) ;
-	if (WEXITSTATUS(i) == 0) fprintf (stderr,"\n");
-	exit (0);
-    }
-
     /*
      * You can suspend, terminate, etc. the parent. We will keep on
      * working in background...
@@ -522,13 +717,13 @@ int main (int argc, char *argv[])
     signal(SIGHUP,SIG_IGN);
     signal(SIGINT,SIG_IGN);
     signal(SIGTERM,SIG_IGN);
-#else
+#endif
+
     if (compat && force)	str="relocating lead-out";
     else if (blank)		str="blanking";
     else			str="formatting";
     if (gui)			gui_action=str;
     else			fprintf (stderr,"* %s .",str);
-#endif
 
     *progress = 1;
 
@@ -575,16 +770,36 @@ int main (int argc, char *argv[])
     }
     else if (mmc_profile==0x12)	// DVD-RAM
     {	cmd[0] = 0x04;		// FORMAT UNIT
-	if ((formats[i+4+4]>>2) == 0x01)
-	    formats[i+1] = 0,
-	    cmd[1] = 0x11;	// "FmtData"|"Format Code"
-	else if (full)
-	    formats[i+1] = 0x82,// "FOV"|"IMMED"
-	    cmd[1] = 0x11;	// "FmtData"|"Format Code"
-	else
-	    formats[i+1] = 0xA2,// "FOV"|"DCRT"|"IMMED"
-	    cmd[1] = 0x19;	// "FmtData"|"CmpLst"|"Format Code"
+	cmd[1] = 0x11;		// "FmtData"|"Format Code"
 	cmd[5] = 0;
+	formats[i+1] = 0x82;	// "FOV"|"IMMED"
+	if ((formats[i+4+4]>>2) != 0x01 && !full)
+	    formats[i+1] |= 0x20,// |"DCRT"
+	    cmd[1] |= 0x08;	// |"CmpLst"
+	if ((err=cmd.transport(WRITE,formats+i,12)))
+	    sperror ("FORMAT UNIT",err), exit(1);
+
+	if (wait_for_unit (cmd,progress)) exit(1);
+    }
+    else if (mmc_profile==0x43)	// BD-RE
+    {	cmd[0] = 0x04;		// FORMAT UNIT
+	cmd[1] = 0x11;		// "FmtData"|"Format Code"
+	cmd[5] = 0;
+	formats[i+1] = 0x82;	// "FOV"|"IMMED"
+	if (full && (formats[i+4+4]>>2)!=0x31)
+	    formats[i+4+4] |= 2;// "Full Certificaton"
+	else if ((formats[i+4+4]>>2)==0x30)
+	    formats[i+4+4] |= 3;// "Quick Certification"
+	if ((err=cmd.transport(WRITE,formats+i,12)))
+	    sperror ("FORMAT UNIT",err), exit(1);
+
+	if (wait_for_unit (cmd,progress)) exit(1);
+    }
+    else if (mmc_profile==0x41)	// BD-R
+    {	cmd[0] = 0x04;		// FORMAT UNIT
+	cmd[1] = 0x11;		// "FmtData"|"Format Code"
+	cmd[5] = 0;
+	formats[i+1] = 0x2;	// IMMED"
 	if ((err=cmd.transport(WRITE,formats+i,12)))
 	    sperror ("FORMAT UNIT",err), exit(1);
 
