@@ -24,18 +24,20 @@
 extern "C" void sigs_mask (int);
 extern "C" void atsignals (void (*)(void));
 
-#define	ONEX 1352	// 1385 * 1000 / 1024
+#define	DVD_1X	1352	// 1385 * 1000 / 1024
+#define BD_1X	4390	// 4495.5 * 1000 / 1024
 
 static int		media_written=0,next_track=1,
-			is_dao=0,quickgrown=0,do_reload=1;
-static unsigned int	stdnap=(256*1000)/ONEX,	// measured in milliseconds
+			is_dao=0,quickgrown=0,do_reload=1,
+			_1x=DVD_1X,_64k=0,bdr_plus_pow=0;
+static unsigned int	stdnap=(256*1000)/DVD_1X,// measured in milliseconds
 			buf_size=512;		// measured in KBs
 static class __velocity__ {
     private:
 	int value;
     public:
-	int operator=(int v)	{ if (v>=ONEX)	stdnap=(buf_size*500)/v;
-				  else		stdnap=(buf_size*500)/ONEX;
+	int operator=(int v)	{ if (v>=_1x)	stdnap=(buf_size*500)/v;
+				  else		stdnap=(buf_size*500)/_1x;
 				  return value = v;
 				}
 	operator int()		{ return value; }
@@ -48,15 +50,19 @@ static void         *ioctl_handle=(void *)-1;
 
 static unsigned int  next_wr_addr=1;	// it starts as boolean
 static unsigned int  dao_blocks=0;
+static unsigned int  cap2kstart=0;
 static unsigned char formats[260],disc_info[32];
 
 extern "C" {
 extern int	dvd_compat,test_write,no_reload,mmc_profile,_argc,
-		wrvfy,no_opc;
+		wrvfy,no_opc,spare;
 extern double	speed_factor;
 extern char	*ioctl_device,**_argv;
 }
 
+extern "C"
+int __1x (void) { return _1x; }
+
 extern "C"
 int fumount (int fd)
 { Scsi_Command cmd;
@@ -64,7 +70,7 @@ int fumount (int fd)
 }
 
 extern "C"
-int media_reload (char *name=NULL,struct stat *sb=NULL)
+int media_reload (char *name=NULL,struct stat *sb=NULL,unsigned int cap2k=0)
 {   if (name==NULL)
     {	Scsi_Command cmd(ioctl_handle);
 
@@ -85,23 +91,33 @@ int media_reload (char *name=NULL,struct stat *sb=NULL)
 	}
 #if !RELOAD_NEVER_NEEDED
 
-	char str[12];
+	char fdstr[12],cap2kstr[12];
 	int n;
 	
 	if ((n=fcntl (ioctl_fd,F_GETFD))<0) n=0;
 	fcntl (ioctl_fd,F_SETFD,n&~FD_CLOEXEC);
 
-	sprintf (str,"%ld",ioctl_fd);
+	sprintf (fdstr,"%ld",ioctl_fd);
+	sprintf (cap2kstr,"%u",cap2kstart);
 	execlp(_argv[0],no_reload<0?"-eject":"-reload",
-			str,ioctl_device,(void *)NULL);
+			fdstr,ioctl_device,cap2kstr,(void *)NULL);
     }
     else
     {
-	{ Scsi_Command cmd;
+	{ Scsi_Command  cmd;
+	  unsigned char c[8];
+	  unsigned int  cap2kend;
 
 	    if (!cmd.associate (name,sb)) return 1;
 
-	    if (cmd.is_reload_needed())
+	    cmd[0] = 0x25;	// READ CAPACITY
+	    cmd[9] = 0;
+	    if (!cmd.transport (READ,c,sizeof(c)))
+		cap2kend = c[0]<<24|c[1]<<16|c[2]<<8|c[3];
+	    else
+		cap2kend = (unsigned int)-1;
+
+	    if (cmd.is_reload_needed(cap2k==cap2kend))
 	    {	fprintf (stderr,"%s: reloading tray\n",name);
 		cmd[0] = 0x1E;		// ALLOW MEDIA REMOVAL
 		cmd[5] = 0;
@@ -157,7 +173,20 @@ int get_mmc_profile (void *fd)
   unsigned char buf[8],inq[128];
   int profile=0,once=1,blank=0,err;
   unsigned int len;
+  
+#if defined(__APPLE__) && defined(__MACH__)
+  // The reason for Mac OS X specific MMCIO appearing here and there
+  // is that raw SCSI requires exclusive access, while I'm not ready
+  // to give it up yet...
 
+    // I don't call INQUIRY, because kernel ensures that
+    // MMCDeviceInterface is attached to an MMC device...
+    if ((err=MMCIO(fd,GetConfiguration,0,0,buf,sizeof(buf))))
+	sperror ("GET CONFIGURATION",err),
+	exit (FATAL_START(errno));
+
+    profile = buf[6]<<8|buf[7];
+#else
     // INQUIRY is considered to be "non-intrusive" in a sense that
     // it won't interfere with any other operation nor clear sense
     // data, which might be important to retain for security reasons.
@@ -209,27 +238,66 @@ int get_mmc_profile (void *fd)
 	    exit (FATAL_START(errno));
 #endif
     } while (once--);
+#endif	// !Mac OS X
 
-    if (profile==0 || (profile&0x30)==0)	// no or non-DVD media...
+    if (profile==0 || (profile&0x70)==0)	// no or non-DVD/BD media...
 	return profile;
 
+    if ((profile&0xF0) == 0x40)
+    { unsigned char header[16];
+
+	_1x = BD_1X;
+	_64k = 1;
+
+	if (profile==0x41)	// BD-R SRM
+	{   // Check for POW feature...
+#if defined(__APPLE__) && defined(__MACH__)
+	    err = MMCIO(fd,GetConfiguration,2,0x38,header,16);
+#else
+	    cmd[0] = 0x46;	// GET CONFIGURATION
+	    cmd[1] = 2;		// ask for the only feature...
+	    cmd[3] = 0x38;	// the "BD-R Pseudo-Overwrite" one
+	    cmd[8] = 16;	// The feature should be there, right?
+	    cmd[9] = 0;
+	    err = cmd.transport (READ,header,16);
+#endif
+	    if (err==0 &&
+		(header[0]<<24|header[1]<<16|header[2]<<8|header[3])>=12 &&
+		(header[8+2]&1)==1)
+		bdr_plus_pow=1;
+	}
+
+    }
+
+#if defined(__APPLE__) && defined(__MACH__)
+    err = MMCIO(fd,ReadDiscInformation,disc_info,sizeof(disc_info));
+#else
     cmd[0] = 0x51;	// READ DISC INFORMATION
     cmd[8] = sizeof(disc_info);
     cmd[9] = 0;
-    if ((err=cmd.transport (READ,disc_info,sizeof(disc_info))))
+    err = cmd.transport (READ,disc_info,sizeof(disc_info));
+#endif
+    if (err)
 	sperror ("READ DISC INFORMATION",err),
 	exit (FATAL_START(errno));
 
     // see if it's blank media
     if ((disc_info[2]&3) == 0)	blank=0x10000;
 
-    if (profile != 0x1A && profile != 0x13 && profile != 0x12)
+    if (profile != 0x1A && profile != 0x2A &&		// DVD+RW
+	profile != 0x13 && profile != 0x12 &&		// DVD-R[AM]
+	profile != 0x43 && !(profile == 0x41 && blank))	// BD-R[E]
 	return blank|profile;
 
+#if defined(__APPLE__) && defined(__MACH__)
+    err = MMCIO(fd,ReadFormatCapacities,formats,12);
+#else
     cmd[0] = 0x23;	// READ FORMAT CAPACITIES
     cmd[8] = 12;
     cmd[9] = 0;
-    if ((err=cmd.transport (READ,formats,12)))
+    err = cmd.transport (READ,formats,12);
+#endif
+    if (err)
 	sperror ("READ FORMAT CAPACITIES",err),
 	exit (FATAL_START(errno));
 
@@ -238,11 +306,16 @@ int get_mmc_profile (void *fd)
 	fprintf (stderr,":-( FORMAT allocaion length isn't sane"),
 	exit (FATAL_START(EINVAL));
 
+#if defined(__APPLE__) && defined(__MACH__)
+    err = MMCIO(fd,ReadFormatCapacities,formats,4+len);
+#else
     cmd[0] = 0x23;	// READ FORMAT CAPACITIES
     cmd[7] = (4+len)>>8;
     cmd[8] = (4+len)&0xFF;
     cmd[9] = 0;
-    if ((err=cmd.transport (READ,formats,4+len)))
+    err = cmd.transport (READ,formats,4+len);
+#endif
+    if (err)
 	sperror ("READ FORMAT CAPACITIES",err),
 	exit (FATAL_START(errno));
 
@@ -256,7 +329,7 @@ int get_mmc_profile (void *fd)
   return blank|profile;
 }
 
-static unsigned int get_2k_capacity (Scsi_Command &cmd)
+static unsigned int get_2k_capacity (Scsi_Command &cmd,void *fd=NULL)
 { unsigned char	buf[32];
   unsigned int	ret=0;
   unsigned int	nwa,free_blocks;
@@ -302,19 +375,27 @@ static unsigned int get_2k_capacity (Scsi_Command &cmd)
 	case 0x16:		// DVD-R Dual Layer Jump
 	case 0x1B:		// DVD+R
 	case 0x2B:		// DVD+R Double Layer
-	    cmd[0] = 0x52;	// READ TRACK INFORMATION
-	    cmd[1] = 1;
-	    cmd[4] = next_track>>8;
-	    cmd[5] = next_track&0xFF;	// last track, set up earlier
-	    cmd[8] = sizeof(buf);
-	    cmd[9] = 0;
-	    if ((err=cmd.transport (READ,buf,sizeof(buf))))
+	case 0x41:		// BD-R SRM
+#if defined(__APPLE__) && defined(__MACH__)
+	    if (fd)
+		err = MMCIO(fd,ReadTrackInformation,1,next_track,buf,sizeof(buf));
+	    else
+#endif
+	    {	cmd[0] = 0x52;	// READ TRACK INFORMATION
+		cmd[1] = 1;
+		cmd[4] = next_track>>8;
+		cmd[5] = next_track&0xFF;	// last track, set up earlier
+		cmd[8] = sizeof(buf);
+		cmd[9] = 0;
+		err = cmd.transport (READ,buf,sizeof(buf));
+	    }
+	    if (err)
 	    {	sperror ("READ TRACK INFORMATION",err);
 		return 0;
 	    }
 
 	    nwa = 0;
-	    if (buf[7]&1)	// NWA_V
+	    if (buf[7]&1 && !bdr_plus_pow)	// NWA_V
 	    {	nwa  = buf[12]<<24;
 		nwa |= buf[13]<<16;
 		nwa |= buf[14]<<8;
@@ -327,6 +408,14 @@ static unsigned int get_2k_capacity (Scsi_Command &cmd)
 	    ret = nwa + free_blocks;
 	    break;
 
+	case 0x43:		// BD-RE
+	    // just pull formatted capacity for now...
+	    ret  = formats[4+0]<<24;
+	    ret |= formats[4+1]<<16;
+	    ret |= formats[4+2]<<8;
+	    ret |= formats[4+3];
+	    break;
+
 	default:
 	    break;
     }
@@ -337,25 +426,25 @@ static unsigned int get_2k_capacity (Scsi_Command &cmd)
 extern "C"
 off64_t get_capacity (void *fd)
 { Scsi_Command	cmd(fd);
-  return (off64_t)get_2k_capacity(cmd)*2048;
+  return (off64_t)get_2k_capacity(cmd,fd)*2048;
 }
 
 extern "C"
-double get_buffer_stats (void *fd)
+float get_buffer_stats (void *fd)
 { Scsi_Command		cmd(fd);
   unsigned char		bcap[12];
-  unsigned int		bsize,bfree;
+  unsigned int		bsize,bfree,err;
 
     cmd[0] = 0x5C;              // READ BUFFER CAPACITY
     cmd[8] = sizeof(bcap);
     cmd[9] = 0;
-    if (cmd.transport (READ,bcap,sizeof(bcap)))
-	return 0.0;
+    if ((err=cmd.transport (READ,bcap,sizeof(bcap))))
+	return -err;
 
     bsize = bcap[4]<<24|bcap[5]<<16|bcap[6]<<8|bcap[7];
     bfree = bcap[8]<<24|bcap[9]<<16|bcap[10]<<8|bcap[11];
 
-  return (1.0 - (double)bfree/(double)bsize);
+  return bsize ? (1.0 - (float)bfree/(float)bsize) : 0.0;
 }
 
 ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
@@ -558,6 +647,7 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
 	    // to disable it! This code is a tribute to users who disregard
 	    // the instructions...
 	sync_cache:
+	    // should I check if lba+nbl is multiple of 64K in BD case?
 	    cmd[0] = 0x35;	// SYNC CACHE
 	    cmd[9] = 0;
 	    if (!cmd.transport())
@@ -601,6 +691,115 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
   return size;
 }
 
+// spare == -1	"format without spare"
+// spare == 0	"format with minimal spare"
+// spare >  0	"format with default descriptor"
+static void bd_r_format (Scsi_Command &cmd)
+{ int err,i,len;
+  unsigned char descr[12],*f;
+
+    if (spare < 0)	return;	// no format -> no spare area:-)
+
+    len = formats[3];
+
+    if (spare == 0)	// locate descriptor with maximum capacity
+    { unsigned int max,cap;
+      int j;
+
+	i = 0;
+	for (max=0,j=8;j<len;j+=8)
+	{   f = formats+4+j;
+	    if ((f[4]>>2) == 0x32)
+	    {	cap = f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+		if (max < cap) max=cap,i=j;
+	    }
+	}
+	if (i==0)
+	    fprintf (stderr,":-( can't locate format 0x32 descriptor\n"),
+	    exit (FATAL_START(errno=EMEDIUMTYPE));
+    }
+    else
+	i = 8;	// grab default descriptor
+
+    memset (descr,0,sizeof(descr));
+    descr[1]=0x02;	// "IMMED" flag
+    descr[3]=0x08;	// "Descriptor Lenght" (LSB)
+    memcpy (descr+4,formats+4+i,5);
+    descr[4+4]&=~3;	// ensure "Format Subtype" is set to SRM+POW 
+
+    f = descr+4;
+    fprintf (stderr,"%s: pre-formatting blank BD-R for %.1fGB...\n",
+		    ioctl_device,(f[0]<<24|f[1]<<16|f[2]<<8|f[3])*2048.0/1e9);
+
+    cmd[0] = 0x04;	// FORMAT UNIT
+    cmd[1] = 0x11;	// "FmtData" and "Format Code"
+    cmd[5] = 0;
+    if ((err=cmd.transport(WRITE,descr,sizeof(descr))))
+	sperror ("FORMAT UNIT",err),
+	exit (FATAL_START(errno));
+
+    wait_for_unit (cmd);
+
+    cmd[0] = 0x35;	// FLUSH CACHE
+    cmd[9] = 0;
+    cmd.transport();
+}
+
+static void bd_re_format (Scsi_Command &cmd)
+{ int err,i,len;
+  unsigned char descr[12],*f;
+
+    len = formats[3];
+
+    if (spare == 0)	// locate descriptor with maximum capacity
+    { unsigned int max,cap;
+      int j;
+
+	i = 0;
+	for (max=0,j=8;j<len;j+=8)
+	{   f = formats+4+j;
+	    if ((f[4]>>2) == 0x30)
+	    {	cap = f[0]<<24|f[1]<<16|f[2]<<8|f[3];
+		if (max < cap) max=cap,i=j;
+	    }
+	}
+	if (i==0)
+	    fprintf (stderr,":-( can't locate format 0x30 descriptor\n"),
+	    exit (FATAL_START(errno=EMEDIUMTYPE));
+    }
+    else if (spare < 0)	// locate descriptor 0x31
+    {	for (i=8;i<len;i+=8)
+	    if ((formats[4+i+4]>>2) == 0x31) break;
+	if (i==len)
+	    fprintf (stderr,":-( can't locate format 0x31 descriptor\n"),
+	    exit (FATAL_START(errno=EMEDIUMTYPE));
+    }
+    else
+	i = 8;	// grab default descriptor
+
+    memset (descr,0,sizeof(descr));
+    descr[1]=0x02;	// "IMMED" flag
+    descr[3]=0x08;	// "Descriptor Lenght" (LSB)
+    memcpy (descr+4,formats+4+i,5);
+
+    f = descr+4;
+    fprintf (stderr,"%s: pre-formatting blank BD-RE for %.1fGB...\n",
+		    ioctl_device,(f[0]<<24|f[1]<<16|f[2]<<8|f[3])*2048.0/1e9);
+
+    cmd[0] = 0x04;	// FORMAT UNIT
+    cmd[1] = 0x11;	// "FmtData" and "Format Code"
+    cmd[5] = 0;
+    if ((err=cmd.transport(WRITE,descr,sizeof(descr))))
+	sperror ("FORMAT UNIT",err),
+	exit(FATAL_START(errno));
+
+    wait_for_unit (cmd);
+
+    cmd[0] = 0x35;	// FLUSH CACHE
+    cmd[9] = 0;
+    cmd.transport();
+}
+
 static void plus_rw_format (Scsi_Command &cmd)
 { int err,i,len;
   unsigned char descr[12];
@@ -705,56 +904,69 @@ int plusminus_r_C_parm (void *fd,char *C_parm)
 	fprintf (stderr,":-( last session is not empty\n"),
 	exit(FATAL_START(EMEDIUMTYPE));
 
-    next_track = disc_info[5]|disc_info[10]<<8;
+    if (mmc_profile==0x41 && bdr_plus_pow)
+	next_track = disc_info[6]|disc_info[11]<<8;
+    else
+	next_track = disc_info[5]|disc_info[10]<<8;
 
+#if defined(__APPLE__) && defined(__MACH__)
+    err = MMCIO(fd,ReadTrackInformation,1,next_track,buf,sizeof(buf));
+#else
     cmd[0] = 0x52;		// READ TRACK INFORMATION
     cmd[1] = 1;
     cmd[4] = next_track>>8;
     cmd[5] = next_track;	// ask for last track
     cmd[8] = sizeof(buf);
     cmd[9] = 0;
-    if ((err=cmd.transport (READ,buf,sizeof(buf))))
+    err = cmd.transport (READ,buf,sizeof(buf));
+#endif
+    if (err)
     {	sperror ("READ TRACK INFORMATION",err);
 	if (ASC(err)==0x24) // hp dvd200i returns 24 if media is full
 	    fprintf (stderr,":-( media must be full already\n");
 	exit (FATAL_START(errno));
     }
 
-#if 0
-    if (buf[7]&1)		// NWA_V
+    if (mmc_profile==0x41 && bdr_plus_pow && buf[7]&1)		// NWA_V
     {	next_session  = buf[12]<<24;
 	next_session |= buf[13]<<16;
 	next_session |= buf[14]<<8;
 	next_session |= buf[15];
     }
     else
-	fprintf (stderr,":-( Next Writable Address is invalid\n"),
-	exit(FATAL_START(EMEDIUMTYPE));
+    {	next_session  = buf[8]<<24;	// Track Start Address
+	next_session |= buf[9]<<16;
+	next_session |= buf[10]<<8;
+	next_session |= buf[11];
+    }
+
+    if (mmc_profile==0x41 && bdr_plus_pow)
+	prev_session=0;
+    else
+    {	//
+	// All manuals say the data is fabricated, presumably implying
+	// that one should use another command. But we stick to this one
+	// because kernel uses this very command to mount multi-session
+	// discs.
+	//
+#if defined(__APPLE__) && defined(__MACH__)
+	err = MMCIO(fd,ReadTableOfContents,0,1,0,buf,12);
 #else
-    next_session  = buf[8]<<24;	// Track Start Address
-    next_session |= buf[9]<<16;
-    next_session |= buf[10]<<8;
-    next_session |= buf[11];
+	cmd[0] = 0x43;		// READ TOC
+	cmd[2] = 1;		// "Session info"
+	cmd[8] = 12;
+	cmd[9] = 0;
+	err = cmd.transport (READ,buf,12);
 #endif
+	if (err)
+	    sperror ("READ SESSION INFO",err),
+	    exit (FATAL_START(errno));
 
-    //
-    // All manuals say the data is fabricated, presumably implying
-    // that one should use another command. But we stick to this one
-    // because kernel uses this very command to mount multi-session
-    // discs.
-    //
-    cmd[0] = 0x43;		// READ TOC
-    cmd[2] = 1;			// "Session info"
-    cmd[8] = 12;
-    cmd[9] = 0;
-    if ((err=cmd.transport (READ,buf,12)))
-	sperror ("READ SESSION INFO",err),
-	exit (FATAL_START(errno));
-
-    prev_session  = buf[8]<<24;
-    prev_session |= buf[9]<<16;
-    prev_session |= buf[10]<<8;
-    prev_session |= buf[11];
+	prev_session  = buf[8]<<24;
+	prev_session |= buf[9]<<16;
+	prev_session |= buf[10]<<8;
+	prev_session |= buf[11];
+    }
 
     sprintf (C_parm,"%d,%d",prev_session+16,next_session);
 
@@ -856,7 +1068,7 @@ static int pull_velocity (Scsi_Command &cmd,unsigned char *d)
   return 0;
 }
 
-static int set_speed_B6h (Scsi_Command &cmd,unsigned int dvddash,
+static int set_speed_B6h (Scsi_Command &cmd,unsigned int dvdplus,
 			  unsigned char *page2A)
 { unsigned int len;
   int err;
@@ -944,17 +1156,17 @@ static int set_speed_B6h (Scsi_Command &cmd,unsigned int dvddash,
 	    } // minv will be treated as 1x
 
 	    // New generation units offer minimum velocities other than 1x
-	    if      (minv>(15*ONEX))	speed_factor /= 16.0;
-	    else if (minv>(11*ONEX))	speed_factor /= 12.0;
-	    else if (minv>(7*ONEX))	speed_factor /= 8.0;
-	    else if (minv>(5*ONEX))	speed_factor /= 6.0;
-	    else if (minv>(3*ONEX))	speed_factor /= 4.0;	
-	    else if (!dvddash && minv>(2*ONEX))
+	    if      (minv>(15*_1x))	speed_factor /= 16.0;
+	    else if (minv>(11*_1x))	speed_factor /= 12.0;
+	    else if (minv>(7*_1x))	speed_factor /= 8.0;
+	    else if (minv>(5*_1x))	speed_factor /= 6.0;
+	    else if (minv>(3*_1x))	speed_factor /= 4.0;	
+	    else if (dvdplus && minv>(2*DVD_1X))
 		// 2.4x is like 1x for DVD+, but it turned out that there're
 		// units, most notably "DVDRW IDE1004," burning DVD+ at
 		// lower speed, so I have to watch out for it...
 		speed_factor /= 2.4;
-	    else if (minv>=(2*ONEX))
+	    else if (minv>=(2*_1x))
 		speed_factor /= 2.0;
 
 	    v0=(int)(speed_factor*minv + 0.5);
@@ -1018,7 +1230,7 @@ static int set_speed_B6h (Scsi_Command &cmd,unsigned int dvddash,
 	{   if (targetv!=velocity)
 	    {	if (errp==0)	// check Page 2A for speed then...
 		{   errp=-1; continue;   }
-		if (velocity==0 || velocity>=ONEX)
+		if (velocity==0 || velocity>=_1x)
 		{   fprintf(stderr,":-? Failed to change write speed: %d->%d\n",
 					(int)velocity,targetv);
 		    if (velocity<targetv)	velocity=targetv;
@@ -1108,7 +1320,7 @@ static int set_speed_BBh (Scsi_Command &cmd,unsigned char *page2A)
 }
 
 static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
-{ unsigned int   plen,hlen,dvddash;
+{ unsigned int   plen,hlen,dvddash=0,dvdplus=0;
   unsigned char  header[16],p32;
   int            err;
   class autofree page2A;
@@ -1116,7 +1328,7 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
     switch (profile)
     {	case 0x1A:	// DVD+RW
 	    p32 = 0;
-	    dvddash = 0;
+	    dvdplus=1;
 	    break;
 	case 0x1B:	// DVD+R
 	    //
@@ -1129,7 +1341,7 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
 	    // the market.
 	    //
 	    p32 = dvd_compat?0x02:0xC0;
-	    dvddash = 0;
+	    dvdplus=1;
 	    break;
 	case 0x13:	// DVD-RW Restricted Overwrite
 	    p32 = 0xC0;	// NB! Test Write in not option here
@@ -1187,7 +1399,6 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
 	    break;
 	default:
 	    p32 = 0;
-	    dvddash = 0;
 	    break;
     }
 
@@ -1209,14 +1420,14 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
 
     // GET PERFORMANCE/SET STREAMING are listed as mandatory, so that
     // this is actually the most likely path...
-    if (set_speed_B6h (cmd,dvddash,page2A) >= 0)
+    if (set_speed_B6h (cmd,dvdplus,page2A) >= 0)
 	goto opc;
 
     if (plen<(hlen+30) || p[1]<(30-2))	// no "Current Write Speed" present
     {	if (dvddash)
 	{   fprintf (stderr,":-( \"Current Write Speed\" descriptor "
 			    "is not present, faking 1x...\n"),
-	    velocity = ONEX;
+	    velocity = _1x;
 	    speed_factor = 0.0;
 	}
 	else
@@ -1227,7 +1438,7 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
 
 	// Some units, most notably NEC and derivatives, were observed
 	// to report CD-R descriptors...
-	if (velocity < ONEX)			// must be bogus
+	if (velocity < DVD_1X)			// must be bogus
 	{   velocity=0;
 	    break;
 	}
@@ -1240,12 +1451,12 @@ static int plusminus_pages_setup (Scsi_Command &cmd,int profile)
 	for (;n--;p_+=4)
 	{   int v=p_[2]<<8|p_[3];
 	    if (v==0) continue;		// Lite-On returns zeros here...
-	    if (v<ONEX)			// must be bogus
+	    if (v<_1x)			// must be bogus
 		velocity=0;
 	    if (v < minv)
 		minv = v;
 	}
-	if (!dvddash && minv>2*ONEX)
+	if (dvdplus && minv>2*_1x)
 	    // See corresponding comment in set_speed_B6h()
 	    speed_factor /= 2.4;
 
@@ -1435,6 +1646,7 @@ static void plus_r_dl_split (Scsi_Command &cmd,off64_t size)
     cmd[7] = 0x20;	// "DVD+R Double Layer Recording Information"
     cmd[9] = sizeof(dvd_20);
     cmd[11] = 0;
+    cmd.timeout(180);	// Lite-on seems to fire off calibration here
     if ((err=cmd.transport(WRITE,dvd_20,sizeof(dvd_20))))
 	sperror ("SEND DVD+R DOUBLE LAYER RECORDING INFORMATION",err),
 	exit (FATAL_START(errno));
@@ -1527,6 +1739,25 @@ void plus_rw_finalize ()
     _exit (errno);
 }
 
+extern "C"
+void bd_re_finalize ()
+{   sigs_mask(SIGS_BLOCK);
+    while (media_written)
+    { Scsi_Command cmd(ioctl_handle);
+
+	fprintf (stderr,"%s: flushing cache\n",ioctl_device);
+	if (flush_cache (cmd))	break;
+
+	media_written = 0;
+	next_wr_addr = 0;
+	errno = 0;
+
+	if (do_reload && media_reload())
+	    perror (":-( unable to reload tray");// not actually reached
+    }
+    _exit (errno);
+}
+
 extern "C"
 void plus_r_finalize ()
 {   sigs_mask(SIGS_BLOCK);
@@ -1674,6 +1905,57 @@ void minus_r_finalize ()
     _exit (errno);
 }
 
+extern "C"
+void bd_r_finalize ()
+{   sigs_mask(SIGS_BLOCK);
+    while (media_written)
+    { Scsi_Command  cmd(ioctl_handle);
+      int           err;
+
+	fprintf (stderr,"%s: flushing cache\n",ioctl_device);
+	if (flush_cache(cmd))	break;
+
+	fprintf (stderr,"%s: closing track\n",ioctl_device);
+	cmd[0] = 0x5B;		// CLOSE TRACK/SESSION
+	cmd[1] = 0x01;		// "IMMED"
+	cmd[2] = 0x01;		// "Close Track"
+	cmd[4] = next_track>>8;
+	cmd[5] = next_track;
+	cmd[9] = 0;
+	if ((err=cmd.transport()))
+	    sperror ("CLOSE TRACK",err);
+
+	if (wait_for_unit (cmd)) break;
+
+	if (!bdr_plus_pow)
+	{   fprintf (stderr,"%s: closing session\n",ioctl_device);
+	    while (1)	// Pioneer DVR-K04RA
+	    {	cmd[0] = 0x5B;		// CLOSE TRACK/SESSION
+		cmd[1] = 0x01;		// "IMMED"
+		cmd[2] = 0x02;		// "Close session"
+		cmd[9] = 0;
+		err = cmd.transport();
+		if (err == 0x20407)		// "OP IN PROGRESS"
+		{   poll (NULL,0,stdnap);
+		    continue;
+		}
+		else if (err)
+		    sperror ("CLOSE SESSION",err);
+		break;
+	    }
+
+	    if (wait_for_unit (cmd)) break;
+ 	}
+ 
+	media_written = 0;
+	errno = 0;
+
+	if (do_reload && media_reload())
+	    perror (":-( unable to reload tray");// not actually reached
+    }
+    _exit (errno);
+}
+
 extern "C"
 void ram_reload ()
 {   sigs_mask(SIGS_BLOCK);
@@ -1705,6 +1987,18 @@ pwrite64_t poor_mans_setup (void *fd,off64_t leadout)
 
     handle_events(cmd);
 
+    // This actually belongs in get_mmc_profile, but the trouble
+    // is that Mac OS X does not implement ReadCapacity method and
+    // therefore we have to do it here, after we went exclusive...
+    if (!(mmc_profile&0x10000))	// non-blank media
+    { unsigned char c[8];
+
+	cmd[0] = 0x25;	// READ CAPACITY
+	cmd[9] = 0;
+	if (!cmd.transport (READ,c,sizeof(c)))
+	    cap2kstart = c[0]<<24|c[1]<<16|c[2]<<8|c[3];
+    }
+
     switch (profile)
     {	case 0x1A:	// DVD+RW
 	    switch (disc_info[7]&3)	// Background formatting
@@ -1785,6 +2079,27 @@ pwrite64_t poor_mans_setup (void *fd,off64_t leadout)
 	    atexit (ram_reload);
 	    atsignals (ram_reload);
 	    break;
+	case 0x41:	// BD-R SRM
+	    if ((disc_info[2]&3) == 0)	// blank
+		bd_r_format (cmd);
+	    // exclusively for speed settings...
+	    plusminus_pages_setup (cmd,profile);
+	    atexit (bd_r_finalize);
+	    if (next_wr_addr)
+	    {	atsignals (no_r_finalize);
+		next_wr_addr=(unsigned int)-1;
+	    }
+	    else atsignals (bd_r_finalize);
+	    break;
+	case 0x43:	// BR-RE
+	    if ((disc_info[2]&3) == 0)	// blank
+		bd_re_format (cmd);
+	    // exclusively for speed settings...
+	    plusminus_pages_setup (cmd,profile);
+	    atexit(bd_re_finalize);
+	    atsignals(bd_re_finalize);
+	    break;
+	case 0x42:	// BD-R RRM is not supported yet...
 	case 0x16:	// DVD-R Dual Layer Jump not supported for reasons
 			// discussed at web-page...
 	default:
@@ -1794,8 +2109,8 @@ pwrite64_t poor_mans_setup (void *fd,off64_t leadout)
     }
 
     if (velocity)
-	fprintf (stderr,"%s: \"Current Write Speed\" is %.1fx1385KBps.\n",
-			ioctl_device,velocity/(double)ONEX);
+	fprintf (stderr,"%s: \"Current Write Speed\" is %.1fx%dKBps.\n",
+			ioctl_device,velocity/(double)_1x,_1x);
 
     if (next_wr_addr==(unsigned int)-1) do
     { unsigned char track[32];
@@ -1837,7 +2152,10 @@ int poor_man_rewritable (void *fd, void *buf)
     if (profile!=0x13 &&	// not DVD-RW Restricted Overwrite
 	profile!=0x12 &&	// nor DVD-RAM
 	profile!=0x1A &&	// nor DVD+RW
-	profile!=0x2A)		// nor DVD+RW Double Layer
+	profile!=0x2A &&	// nor DVD+RW Double Layer
+	!(profile==0x41 && bdr_plus_pow) && // nor BD-R SRM+POW
+	profile!=0x42 &&	// nor BD-R RRM
+	profile!=0x43)		// nor BD-RE
 	return 0;
 
     if (profile==0x13)		// DVD-RW Restricted Overwrite
@@ -1847,8 +2165,9 @@ int poor_man_rewritable (void *fd, void *buf)
 	minus_rw_finalize ();	// with do_reload==0!
 	do_reload=1;
     }
-    else if (profile==0x1A ||	// DVD+RW
-	     profile==0x2A)	// DVD+RW Double Layer
+    else if (profile==0x1A ||		// DVD+RW
+	     profile==0x2A ||		// DVD+RW Double Layer
+	     (profile&=0xF0)==0x40)	// BD-R[E]
     {	fprintf (stderr,"%s: flushing cache\n",ioctl_device);
 	if (flush_cache(cmd))	exit (errno);
     }
